/*
	This program performs various IOCTLs on linux cdrom devices.
	
	Some of the ioctls used (for example, CDROM_DRIVE_STATUS) are
	unavailable in linux 2.2.

	TODO:   Add command line specified times, tracks, etc.
		Add command line option to change cdrom device name.
	
	(c) John Fremlin
		See message_id:	below for dates, etc.
		See AUTHORS.
	
	Under GPL, see file COPYING.
*/


/* Below stolen from <linux/cdrom.h>, kernel 2.3.99pre6
 *	
 *  Copyright (C) 1992         David Giller, rafetmad@oxy.edu
 *                1994, 1995   Eberhard Moenkeberg, emoenke@gwdg.de
 *                1996         David van Leeuwen, david@tm.tno.nl
 *                1997, 1998   Erik Andersen, andersee@debian.org
 *                1998-2000    Jens Axboe, axboe@suse.de
 *
 *		[Certainly not (C) John Fremlin in any way at all]
 */
	

#define EDRIVE_CANT_DO_THIS  EOPNOTSUPP

#define CDROMPAUSE		0x5301 /* Pause Audio Operation */ 
#define CDROMRESUME		0x5302 /* Resume paused Audio Operation */
#define CDROMPLAYMSF		0x5303 /* Play Audio MSF (struct cdrom_msf) */
#define CDROMPLAYTRKIND		0x5304 /* Play Audio Track/index 
                                           (struct cdrom_ti) */
#define CDROMREADTOCHDR		0x5305 /* Read TOC header 
                                           (struct cdrom_tochdr) */
#define CDROMREADTOCENTRY	0x5306 /* Read TOC entry 
                                           (struct cdrom_tocentry) */
#define CDROMSTOP		0x5307 /* Stop the cdrom drive */
#define CDROMSTART		0x5308 /* Start the cdrom drive */
#define CDROMEJECT		0x5309 /* Ejects the cdrom media */
#define CDROMVOLCTRL		0x530a /* Control output volume 
                                           (struct cdrom_volctrl) */
#define CDROMSUBCHNL		0x530b /* Read subchannel data 
                                           (struct cdrom_subchnl) */
#define CDROMREADMODE2		0x530c /* Read CDROM mode 2 data (2336 Bytes) 
                                           (struct cdrom_read) */
#define CDROMREADMODE1		0x530d /* Read CDROM mode 1 data (2048 Bytes)
                                           (struct cdrom_read) */
#define CDROMREADAUDIO		0x530e /* (struct cdrom_read_audio) */
#define CDROMEJECT_SW		0x530f /* enable(1)/disable(0) auto-ejecting */
#define CDROMMULTISESSION	0x5310 /* Obtain the start-of-last-session 
                                           address of multi session disks 
                                           (struct cdrom_multisession) */
#define CDROM_GET_MCN		0x5311 /* Obtain the "Universal Product Code" 
                                           if available (struct cdrom_mcn) */
#define CDROM_GET_UPC		CDROM_GET_MCN  /* This one is depricated, 
                                          but here anyway for compatability */
#define CDROMRESET		0x5312 /* hard-reset the drive */
#define CDROMVOLREAD		0x5313 /* Get the drive's volume setting 
                                          (struct cdrom_volctrl) */
#define CDROMREADRAW		0x5314	/* read data in raw mode (2352 Bytes) */
#define CDROMREADCOOKED		0x5315	/* read data in cooked mode */
#define CDROMSEEK		0x5316  /* seek msf address */
#define CDROMPLAYBLK		0x5317	/* (struct cdrom_blk) */
#define CDROMREADALL		0x5318	/* read all 2646 bytes */
#define CDROMGETSPINDOWN        0x531d
#define CDROMSETSPINDOWN        0x531e
#define CDROMCLOSETRAY		0x5319	/* pendant of CDROMEJECT */
#define CDROM_SET_OPTIONS	0x5320  /* Set behavior options */
#define CDROM_CLEAR_OPTIONS	0x5321  /* Clear behavior options */
#define CDROM_SELECT_SPEED	0x5322  /* Set the CD-ROM speed */
#define CDROM_SELECT_DISC	0x5323  /* Select disc (for juke-boxes) */
#define CDROM_MEDIA_CHANGED	0x5325  /* Check is media changed  */
#define CDROM_DRIVE_STATUS	0x5326  /* Get tray position, etc. */
#define CDROM_DISC_STATUS	0x5327  /* Get disc type, etc. */
#define CDROM_CHANGER_NSLOTS    0x5328  /* Get number of slots */
#define CDROM_LOCKDOOR		0x5329  /* lock or unlock door */
#define CDROM_DEBUG		0x5330	/* Turn debug messages on/off */
#define CDROM_GET_CAPABILITY	0x5331	/* get capabilities */
#define CDROMAUDIOBUFSIZ        0x5382	/* set the audio buffer size */
#define DVD_READ_STRUCT		0x5390  /* Read structure */
#define DVD_WRITE_STRUCT	0x5391  /* Write structure */
#define DVD_AUTH		0x5392  /* Authentication */
#define CDROM_SEND_PACKET	0x5393	/* send a packet to the drive */
#define CDROM_NEXT_WRITABLE	0x5394	/* get next writable block */
#define CDROM_LAST_WRITTEN	0x5395	/* get last block written on disc */
#define CDS_NO_INFO		0	/* if not implemented */
#define CDS_NO_DISC		1
#define CDS_TRAY_OPEN		2
#define CDS_DRIVE_NOT_READY	3
#define CDS_DISC_OK		4
#define CDS_AUDIO		100
#define CDS_DATA_1		101
#define CDS_DATA_2		102
#define CDS_XA_2_1		103
#define CDS_XA_2_2		104
#define CDS_MIXED		105
#define	CDROM_LBA 0x01 /* "logical block": first frame is #0 */
#define	CDROM_MSF 0x02 /* "minute-second-frame": binary, not bcd here! */
#define	CDROM_AUDIO_INVALID	0x00	/* audio status not supported */
#define	CDROM_AUDIO_PLAY	0x11	/* audio play operation in progress */
#define	CDROM_AUDIO_PAUSED	0x12	/* audio play operation paused */
#define	CDROM_AUDIO_COMPLETED	0x13	/* audio play successfully completed */
#define	CDROM_AUDIO_ERROR	0x14	/* audio play stopped due to error */
#define	CDROM_AUDIO_NO_STATUS	0x15	/* no current audio status to return */

#define	CDROM_LEADOUT		0xAA
			
/* End inclusion from linux/cdrom.h */

#include "config.h"
	
/* 
	%edi contains the fd of the cd device or -1
	%ebp contains the filename of the cd device
	%gs contains an OS error return (I don't want a .data rw segment) 
	(i.e. callee save on these registers)

	Direction flag is cleared by default
*/
			
#define NOT_OPEN	0xffffffff
#define STDOUT		1
#define	commandlength	16
			
	.text

	.globl	_start
_start:
	popl %edx		 /* argc */
	popl %esi		 /* argv[0] */

	xorl %eax,%eax
	movw %ax,%gs
	movl $NOT_OPEN,%edi
	movl $default_cdrom,%ebp
	
	cld

remove_path:
	movl %esi, %ebx /* %ebx will point to character after last "/" */

find_next_slash:	
	lodsb
	cmpb $'/',%al
	jz remove_path

	testb %al,%al
	jnz find_next_slash

	movb (%ebx),%al
		
	testb %al,%al
	jz parse_next_command
	
	/* Were argv to end in a /, this would be the outcome.  This
	is unlikely to happen <grin>, therefore I haven't bothered
	writing an error description */

	movl %ebx,%eax
remove_any_cd_prefix:		
	cmpw $0x6463,(%eax)	 /* 0x64 0x63 == 'cd' */
	jnz lookup_command
	
	incl %eax
	incl %eax
	jmp lookup_command
		
parse_next_command:	
	
        popl %eax		 /* argv[i] */
	testl %eax,%eax
	jz finished	

lookup_command:
		
	movl $command_list,%esi
	movl (%eax),%eax	
		
lookup_loop:
	/*
	 pointer to command string is in %eax
	 current command to check is pointed to by %esi
	*/
	cmpl $0,(%esi)
	je bad_command

	movl (%esi),%ebx
	cmpw %ax,(%ebx)
	
	jz found_command

	addl $commandlength,%esi

	jmp lookup_loop
	
finished: /* IN:	%gs:	return value to shell */
	xorl %ebx,%ebx
	movw %gs,%bx
	xorl %eax,%eax		
	incl %eax		 /* %eax now == 1 */
	int $0x80

bad_drive_status:
	pushl %esi
	movl $error_drive_status,%esi
	call print
	pop %esi
	ret
	
no_disc_status:
	pushl %esi
	movl $error_no_disc_status,%esi
	call print
	pop %esi
	ret
	
bad_disc_status:	
	pushl %esi
	movl $error_disc_status,%esi
	call print
	pop %esi
	ret

bad_disc:
	movl $error_disc,%esi
	call print
	ret	
			
bad_command:
	movw %ds,%ax
	movw %ax,%gs

	pushl %esi
	movl $error_command,%esi
	call print
	popl %esi
	jmp parse_next_command

bad_ioctl:
	
	pushl %eax

	
	movw %ds,%ax
	movw %ax,%gs

	movl $error_ioctl,%esi
	call print
	movl (%esp),%eax
	call print_number
	call print_nl
	popl %eax
	ret

bad_audio_status:
	movl $error_audio_status,%esi
	call print
	xorl %eax,%eax
	xorl %ebx,%ebx
	ret

found_command:
	
do_command:
	pushl $parse_next_command /* when ebx rets, it gets here */

	lodsl
	lodsl
	movl %eax,%ebx
		 /* %ebx now contains pointer to handler function */
	lodsl

	jmp *%ebx

ensure_open_cd_device:
	cmpl $NOT_OPEN,%edi
	jnz ret
	
open_cd_device:
	/*
	 IN: %ebp:	pointer to device filename
	 OUT: %edi:	filedescriptor
	 CLOBBERED: %ebx, %eax
	*/

	movl $5, %eax		 /* 5 == open */
	movl %ebp, %ebx
	movl $04000, %ecx	 /* 04000 == O_NONBLOCK */

	int $0x80

	cmp $0xfffff001,%eax  /* if a syscall returns [-1...-4096] it is an error */
	jae bad_open

	movl	%eax,%edi
	
	ret

bad_open:
	movl $error_open, %esi
	call print
	movw %ds,%ax
	movw %ax,%gs
	jmp finished
		
print:	
	/*			
	 IN:		%esi:	pointer to string to print 
	 CLOBBERED:	%edx, %ecx, %ebx, %esi
	*/
	
	xorl %edx,%edx
	decl %edx

	push %esi
		
count_string:
	incl %edx
	lodsb
	test %al,%al
	jnz count_string

	/*
	 now %edx contains the number of characters in the string.
	*/
	pop %ecx
	
syscall_write:	
	movl $4,%eax		 /* 4 == write */
	movl $STDOUT,%ebx

	int $0x80
	
	ret
print_zero:
	movl $ascii_zero,%ecx
	movl $1,%edx
	jmp syscall_write
	
print_number:
	/* IN:	%eax:	 number */
	testl %eax,%eax
	jz print_zero
	
	push %edi
	
	movl %esp,%edi
	decl %edi
	
	movl %eax,%ebx
	
	subl $12,%esp /* allocate space for number */
	
	movl $10,%esi
	
	testl $0x80000000,%eax /* is it negative? */
	jz print_number_positive
	
print_number_negative:
	pushl %eax
	movl $ascii_hyphen,%ecx
	movl $1,%edx
	call syscall_write
	popl %eax
		
	notl %eax
	incl %eax
	movl %eax,%ebx

print_number_positive:	
	std

	xorl %ecx,%ecx
	decl %ecx
	
	
divide:
	xorl %edx,%edx
	divl %esi
	/* remainder to %edx */

	testl %edx,%edx
	jnz next_digit

	testl %eax,%eax
	jz terminate_string
	
next_digit:	
	movl %eax,%ebx
	movl %edx,%eax
	
	addb $'0',%al
	
	stosb

	movl %ebx,%eax	

	loop divide

terminate_string:	
	cld
	incl %edi
	
	not %ecx

	movl %ecx, %edx
	movl %edi, %ecx

	call syscall_write

	addl $12,%esp
	
	popl %edi

	ret

print_nl:

	pushl $0xa
	movl %esp,%ecx
	
	xorl %edx,%edx
	inc %edx
	
	call syscall_write

	popl %eax
	ret
		
simple_command:
/*
	 IN:	%edi:	contains fd
	 IN:	%eax:	IOCTL number
	 OUT:	%eax:	error status
	 CLOBBERED:	%ebx,%ecx
*/
	
	movl %eax,%ecx

cd_ioctl: 
	/* IN: %edi:	fd
	   IN: %ecx:	IOCTL number
	   IN: %edx:	IOCTL arg
	   OUT:	%eax:	result
	   CLOBBERED:	anything+everything
	*/	
	cmpl $NOT_OPEN,%edi
	jnz do_cd_ioctl

	push %ecx
	push %edx
	call open_cd_device
	pop %edx
	pop %ecx
	
do_cd_ioctl:	
	movl %edi,%ebx
	movl $54,%eax		/* 54 == IOCTL */

	int $0x80
	
	cmp $0xfffff001,%eax  /* if a syscall returns [-1...-4096] it is an error */
	jae bad_ioctl

ret:	ret

drive_status:
	
	movl $CDROM_DRIVE_STATUS,%ecx
	call cd_ioctl
	
	cmpl $CDS_DISC_OK,%eax
	ja bad_drive_status

print_drive_status:	
	movl drive_status_table(,%eax,4),%esi

	pushl %eax
	call print
	popl %eax
	
	ret

disc_status:	
	movl $CDROM_DISC_STATUS,%ecx
	call cd_ioctl

	cmpl $CDS_AUDIO,%eax
	jae check_disc_status

	cmpl $CDS_NO_INFO,%eax
	je no_disc_status
	
	cmpl $CDS_DISC_OK,%eax
	ja bad_disc_status

	jmp print_drive_status
		
check_disc_status:
	cmpl $CDS_MIXED,%eax
	ja bad_disc_status
	
print_disc_status:
	pushl %eax
	subl $CDS_AUDIO,%eax
	movl disc_status_table(,%eax,4),%esi

	call print
	popl %eax
	ret

toc:	call print_toc_header
	xorl %ebx,%ebx
	decl %ebx
	cmpb $0xff,%al
	jne do_print_toc_range
	ret
	
print_toc:
	/* OUT:	 %al:	lowest audio track
	OUT:	%ah:	highest audio track
	*/
	pushl %ebx
	call print_toc_header
	popl %ebx
	
	cmpb $0xff,%al
	jne do_print_toc_range
	ret
	
print_toc_range:
	/* IN:	 %al:	start track
	   IN:	%ah:	end track
	OUT:	 %al:	lowest audio track
	OUT:	%ah:	highest audio track
	*/

	pushl %eax
	call ensure_open_cd_device /* we'll be messing with %ebp */
	popl %eax
	
do_print_toc_range:	
	pushl %ebp
	xorl %ebp,%ebp
	decl %ebp
		
	subl $0x10,%esp
	movb $CDROM_MSF, 2(%esp)
	
print_track: /* the lowest known audio track is in the lowest bytes of %ebp */
	/* IN:	%ebp:	lowest known audio track
	IN:	%ah:	last track to do
	IN:	%al:	track to start at
	*/
	
	cmpb %ah,%al
	ja print_toc_ret
		
	movb %al,(%esp)
	
	movl %esp,%edx

	pushl %eax
	movl $CDROMREADTOCENTRY,%ecx
	call do_cd_ioctl

	movl $message_track, %esi
	call print
	
	xorl %eax,%eax
	movb 4(%esp),%al
	call print_number
		
	movb 5(%esp),%al
	testb $ 0x40,%al
	jnz print_track_data

print_track_audio:
	movl $message_track_audio,%esi
	
	movl (%esp),%eax
	movl %ebp, %ecx
	
	cmpb %al,%cl
	jb do_print_track_audio
	ja got_earlier_track

	cmpl %eax,%ecx
	jbe do_print_track_audio
		
got_earlier_track:
	movl %eax,%ebp
		
do_print_track_audio:	
	jmp print_track_type

got_track_and_ret:
	popl %eax
	pushl $print_toc_ret	
	jmp print_track_type_func
	
print_track_data:
	movl $message_track_data,%esi

print_track_type:
	pushl $print_track_loop
	
print_track_type_func:	
	pushl $print_nl
	jmp print

print_track_loop:	
	popl %eax
	
	incb %al
	jmp print_track
	
print_toc_ret:	
	addl $0x10,%esp
	movl %ebp,%ecx
	movb %cl,%al
	popl %ebp
	ret
	
print_toc_header: /* OUT: %eax.  %al == first track, %ah == last, if ffff invalid */
	pushl $0x0000ffff
	movl $CDROMREADTOCHDR,%ecx
	movl %esp,%edx
	call cd_ioctl
	
	movl $message_start_track,%esi
	call print

	xorl %eax,%eax
	movb (%esp),%al
	
	pushl $print_end_track
	pushl $print_nl
	jmp print_number

print_end_track:
	movl $message_end_track,%esi
	call print

	xorl %eax,%eax
	movb 1(%esp),%al
	
	pushl $print_toc_header_ret
	pushl $print_nl
	jmp print_number

print_toc_header_ret:	
	popl %eax
	ret
	
info:	call read_sub_channel

	testl %eax,%eax
	jz ret

	pushl %eax
	
	movl $message_playing_track,%esi
	call print

	popl %eax
	andl $0xff,%eax
	call print_number

	jmp print_nl
	
resume_play:
	movl $CDROMRESUME,%ecx
	jmp cd_ioctl
possibly_start_play:
	/* IN:	%edx:	if 1 start play */
	decl %edx
	jz play
	ret
			
intelli_start_play: /* IN:	%al:	CDS_AUDIO or CDS_MIXED */
	cmpb $CDS_MIXED,%al
	jz start_play_find_start

	/* Cannot use CDROM_LEADOUT because that confuses ide-scsi
	* emulated CDROMs */	
	call toc
	
start_play: /* IN:	%al:	track from which to start playing
	IN: %ah:	last track to play */	
	xorl %ebx,%ebx
	movb %ah,%bl
	pushw %bx
	movb %al,%bl
	pushw %bx

	movl %esp,%edx

	movl $CDROMPLAYTRKIND,%ecx
	call cd_ioctl
	popl %ebx
	ret

start_play_find_start:
	xorl %ebx,%ebx	
	call print_toc
	
	cmpl $-1,%eax
	jne start_play
	ret
			
play:		
make_disc_ready_to_play:
	call drive_status
	
	cmpb $CDS_DRIVE_NOT_READY,%al
	je inform_try_to_close_tray
	
	cmpb $CDS_TRAY_OPEN,%al /* old status number before kernel 2.4.1 */
	je try_to_close_tray
	
	cmpb $CDS_NO_DISC,%al	/* new status number in kernel 2.4.1 */
	je try_to_close_tray

check_ready_to_play:
	cmpb $CDS_NO_DISC,%al
	je ret

	call disc_status

	cmpb $CDS_AUDIO,%al
	je audio_disc
	
	cmpb $CDS_MIXED,%al
	jne bad_disc 
		
audio_disc:
	pushl %eax
	
	call read_sub_channel

	popl %eax

	cmpb $CDROM_AUDIO_PLAY, %bl
	je ret

	cmpb $CDROM_AUDIO_PAUSED, %bl
	jz resume_play
	
	jmp intelli_start_play
	
inform_try_to_close_tray:
	movl $message_close_tray,%esi
	call print

try_to_close_tray:
	movl $CDROMCLOSETRAY,%ecx
	call cd_ioctl

	call drive_status
	jmp check_ready_to_play

read_sub_channel:
	/* OUT:	%ebx:	audio status - CDROM_AUDIO_PLAY
	   OUT:	%al:	track or 0 if not playing
	   OUT:	%ah:	index or 0 if not playing
	*/	
	subl $0x10, %esp /* allocate 16 bytes for cdrom_subchnl */

	movb $CDROM_MSF,(%esp)
	movl %esp,%edx
	
	movl $CDROMSUBCHNL,%ecx
	call cd_ioctl	

	cmp $0xfffff001,%eax
	jae bad_sub_channel

	
	/* no error */
	xorl %eax,%eax
	movb 1(%esp),%al

	movl %eax,%ebx
	
	pushl $get_current_track
	
print_audio_status:	
	subb $CDROM_AUDIO_PLAY,%al
	cmpb $ CDROM_AUDIO_NO_STATUS - CDROM_AUDIO_PLAY,%al
	ja bad_audio_status

	movl audio_status_table(,%eax,4),%esi

	pushl %ebx
	call print
	popl %ebx

	ret
	
get_current_track: 
	/* IN:	%bl:	if %bl == CDROM_AUDIO_PLAY or CDROM_AUDIO_PAUSED 
			then print track 
	*/
	cmpb $CDROM_AUDIO_PAUSED,%bl
	ja replace_stack
	cmpb $CDROM_AUDIO_PLAY,%bl
	jb replace_stack

	movl 3(%esp),%eax /* %al = track, %ah = index */
	andl $0xffff,%eax
	jmp replace_stack
bad_sub_channel:
	xorl %eax,%eax
replace_stack:
	addl $0x10, %esp

	ret
	
track_offset_play:
	push %eax /* track offset */

	call read_sub_channel
	andl $0xff,%eax
	jz track_offset_play_ret

	popl %ebx
	addl %ebx,%eax

	pushl %eax
	call print_toc_header
	popl %ebx
	movb %bl,%al

	testb %al,%al
	jnz track_offset_check_upper

	movb %ah,%al
	jmp track_offset_find_track 

track_offset_check_upper:
	cmpb %ah,%al
	jbe track_offset_find_track
	movb $1,%al

track_offset_find_track:
	xorl %ebx,%ebx	
	call print_toc_range
	testl %eax,%eax
	jz ret
	
	jmp start_play
track_offset_play_ret:
	pop %eax
	ret
		
autoeject:
	movl %eax,%edx
	movl $CDROMEJECT_SW,%ecx

	jmp cd_ioctl

	
lock:	
	movl %eax,%edx
	movl $CDROM_LOCKDOOR,%ecx

	jmp cd_ioctl

help:
	movl $message_id,%esi
	call print

	movl $message_syntax,%esi
	call print
	
	movl $command_list,%esi

	pushl %edi
help_print_command:
	lodsl

	testl %eax,%eax
	jz help_done
		
	movl %esi,%edi
	pushl %eax
	
	movl $message_command_strt,%esi
	call print

	popl %esi
	
	call print
	movl $message_command_sep,%esi
	call print
	movl %edi,%esi
	lodsl
	lodsl
	lodsl

	movl %esi,%edi
	movl %eax,%esi
	call print
	movl $message_command_term,%esi
	call print
	movl %edi,%esi

	jmp help_print_command
help_done:		
	popl %edi
	ret
		
device:
	movl $message_device,%esi
	call print
	movl %ebp,%esi
	pushl $print_nl
	jmp print

	.section        .rodata

message_id:
	.ascii "cdd -- cdrom device controller for linux\n"
	.ascii "       with "
	.ascii PACKAGE
	.ascii " "
	.ascii VERSION
	.ascii " (C) 2000,2002 John Fremlin\n"
	.string "       distibuted under GPL\n"

message_syntax:
	.ascii "\n\nSyntax:	cdd [command1] [command2] [command3] [command...]\n\n"
	.ascii "Note that the name by which the program is called is considered\n"
	.ascii "to be a command (with any path and \"cd\" prefix removed)\n\n"
	.ascii "Only the first two letters (and case) of a command are signifcant\n\n"
	.string "Valid commands:\n"

message_command_strt:	
	.string "\t"	
message_command_sep:
	.string ": "
message_command_term:
	.string "\n"
			
message_device:	
	.string "CD-ROM device filename: "		
message_close_tray:
	.string "\t[trying to close tray]\n"
message_playing_track:
	.string "Playing track "
message_start_track:
	.string "First track is "
message_end_track:	
	.string "Last track is "	
message_track:
	.string "Track "
message_track_data:	
	.string " (data)"
message_track_audio:	
	.string " (audio)"	
error_open:	
	.string "Could not open CD drive device\n"
error_command:	
	.string "Bad command line argument; try \"help\" (yes, no \"--\" prefix)\n"  
error_ioctl:
	.string "Error performing IOCTL on CD drive: "
error_drive_status:
	.string "Bad drive status from drive\n"	
error_disc_status:	
	.string "Bad disc status from drive\n"
error_no_disc_status:
	.string "Disc status not available\n"
error_disc:
	.string	"Drive does not contain an audio disc\n"
error_audio_status:
	.string "Invalid audio status or audio status not supported\n"
					
doc_again:		.string		"start playing the current track again"	
doc_auto:		.string		"set the CD drive driver to automatically eject and withdraw the tray"	
doc_start:		.string		"start playing the first audio track"
doc_closetray:		.string		"withdraw the tray"
doc_resume:		.string		"resume playing audio if it was paused"
doc_nul:		.string		"start playing if no commands (so program can be called as cdd)"
doc_disc_status:	.string		"print information about the kind of disc in the drive"
doc_drive_status:	.string		"print information about the tray position"
doc_eject:		.string		"spit out the tray"
doc_info:		.string		"print current track, if any"
doc_lockdoor:		.string		"lock the door (so that it can only be opened by software)"
doc_manual:		.string		"set the CD drive driver not to automatically control the tray"
doc_play:		.string		"try to start playing (tries lots of IOCTLs)"
doc_pause:		.string		"pause audio playing"
doc_reset:		.string		"ask the CD drive driver to hardware reset the drive"
doc_skip:		.string		"start playing the next audio track"
doc_toc:		.string		"print the CDs track listing"	
doc_spinup:		.string		"spinup the CD (is this a NOP?)"
doc_stop:		.string		"stop audio playing, spindown the CD"
doc_sub_channel:	.string		"one line message about what the drive is doing"
doc_unlockdoor:		.string		"unlock the door so that it can be opened normally"
doc_help:		.string		"print command line help"
doc_device:		.string		"print cdrom device used"
doc_previous:		.string		"play the track before the current one"	
		
cmd_again:		.string		"again"	
cmd_auto:		.string		"auto"	
cmd_start:		.string		"begin"
cmd_closetray:		.string		"closetray"
cmd_resume:		.string		"continue"
cmd_nul:		.byte		'd'
			.byte		0
cmd_device:		.string		"device"	
cmd_disc_status:	.string		"disc-status"
cmd_drive_status:	.string		"drive-status"
cmd_eject:		.string		"eject"
cmd_info:		.string		"info"
cmd_help:		.string		"help"	
cmd_lockdoor:		.string		"lock"
cmd_manual:		.string		"manual"
cmd_skip:		.string		"next"
cmd_play:		.string		"play"
cmd_pause:		.string		"pause"
cmd_previous:		.string		"previous"	
cmd_reset:		.string		"reset"
cmd_toc:		.string		"toc"	
cmd_spinup:		.string		"spinup"
cmd_stop:		.string		"stop"
cmd_sub_channel:	.string		"sub-channel"
cmd_unlockdoor:		.string		"unlock"
		
default_cdrom:	
	.string		"/dev/cdrom"	
	
command_list:	
	.long		cmd_eject
	.long		simple_command
	.long		CDROMEJECT
	.long		doc_eject
		
	.long		cmd_closetray
	.long		simple_command
	.long		CDROMCLOSETRAY
	.long		doc_closetray
	
	.long		cmd_stop
	.long		simple_command
	.long		CDROMSTOP
	.long		doc_stop

	.long		cmd_resume
	.long		simple_command
	.long		CDROMRESUME
	.long		doc_resume

		
	.long		cmd_pause
	.long		simple_command
	.long		CDROMPAUSE
	.long		doc_pause
	
	.long		cmd_reset
	.long		simple_command
	.long		CDROMRESET
	.long		doc_reset

	.long		cmd_spinup
	.long		simple_command
	.long		CDROMSTART
	.long		doc_spinup

	.long		cmd_auto
	.long		autoeject
	.long		1
	.long		doc_auto

	.long		cmd_manual
	.long		autoeject
	.long		0
	.long		doc_manual

	.long		cmd_lockdoor
	.long		lock
	.long		1
	.long		doc_lockdoor

	.long		cmd_unlockdoor
	.long		lock
	.long		0
	.long		doc_unlockdoor
	
	.long		cmd_play
	.long		play
	.long		0
	.long		doc_play

	.long		cmd_skip
	.long		track_offset_play
	.long		1
	.long		doc_skip
	
	.long		cmd_again
	.long		track_offset_play
	.long		0
	.long		doc_again

	.long		cmd_previous
	.long		track_offset_play
	.long		-1
	.long		doc_previous
	
	.long		cmd_info
	.long		info
	.long		0
	.long		doc_info
	
	.long		cmd_drive_status
	.long		drive_status
	.long		0
	.long		doc_drive_status

	.long		cmd_disc_status
	.long		disc_status
	.long		0
	.long		doc_disc_status

	.long		cmd_sub_channel
	.long		read_sub_channel
	.long		0
	.long		doc_sub_channel

	.long		cmd_start
	.long		intelli_start_play
	.long		0
	.long		doc_start

	.long		cmd_toc
	.long		toc
	.long		0
	.long		doc_toc

	.long		cmd_nul
	.long		possibly_start_play
	.long		0
	.long		doc_nul

	.long		cmd_help
	.long		help
	.long		0
	.long		doc_help

	.long		cmd_device
	.long		device
	.long		0
	.long		doc_device
	
	.long		0
disc_status_table:
	.long 1f
	.long 2f
	.long 3f
	.long 4f
	.long 5f
	.long 6f

1:	.string "CD-DA (audio) disc\n"
2:	.string "Data mode 1 form 1 (yellow book) disc\n"
3:	.string "Data mode 1 form 2 (yellow book) disc\n"
4:	.string "XA mode 2 form 1 (green book) disc\n"
5:	.string "XA mode 2 form 2 (green book) disc\n"
6:	.string	"Mixed format disc\n"
	
drive_status_table:
	.long	1f
	.long	2f
	.long	3f
	.long	4f
	.long	5f
1:		
	.string		"Drive status not supported\n"
2:	
	.string		"No disc in drive\n"
3:	
	.string		"Tray is open\n"
4:	
	.string		"Drive not ready. Closing tray?\n"
5:		
	.string		"Disc in drive\n"

	
audio_status_table:	
	.long 1f
	.long 2f
	.long 3f
	.long 4f
	.long 5f
	
1:	.string "Playing audio\n"
2:	.string "Playing audio (paused)\n"
3:	.string "Finished playing audio\n"
4:	.string "Playing of audio was stopped due to error\n"
5:	.string "Not playing audio\n"

ascii_zero:
	.byte	'0'
ascii_hyphen:	
	.byte	'-'
